// Some of these files already include the other glh headers
// These files must be included in every cpp or h file that needs to use them
//#include <nvgl\glh_extensions.h>
//#include <nvgl\glh_array.h>
//#include <nvgl\glh_combiners.h>
//#include <nvgl\glh_mipmaps.h>

#include "oglapp.h"

// Static so g_ogl cant be used outside of this file
static COGLApplication *g_ogl = NULL;

LRESULT CALLBACK WndProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
	return g_ogl->MsgProc( hWnd, msg, wParam, lParam );
}

COGLApplication::COGLApplication()
{
	g_ogl = this;	// Set the opengl class global to point to this class

	m_bActive         = false;
	m_bFrameMoving    = true;		// Do not start out paused
	m_bReady          = false;		// Not ready untill programmer says its ready
	m_bShowCursor     = true;		// Default to show cursor
	m_bSingleStep     = false;		// Default to running animation
	m_bUseDepthBuffer = true;		// Default to using the depth buffer
	m_bWindowed       = true;		// Default to windowed mode

	m_bRegCombinReady = false;
	m_bTexShaderReady = false;
	m_bVtxProgReady   = false;

	// Remember to chagne this to WS_POPUP when going into fullscreen mode!
	m_dwWindowStyle   = WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;

	m_fElapsedTime     = 0.0f;
	m_fFramesPerSecond = 0.0f;
	m_fTime            = 0;//timeGetTime();	// Change this to QueryPerformanceCounter

	m_hDC       = NULL;
	m_hRC       = NULL;
	m_hWnd      = NULL;
	m_hWndFocus = NULL;

	m_iWindowWidth  = 640;	// ...	
	m_iWindowHeight = 480;	// Make sure and set these two when the resolution chages
	m_iMinDepthBits   = 16;
	m_iMinStencilBits = 0;

	m_ucCurrentResolution = 0x00000003;	// Bit 3 is 640x480

	m_wmCurrentWindowMode = WM_WINDOW;

	m_strWindowTitle = new char[256];

	m_bVidRestart = false;
}
	

/*
===============
DisplayErrorMsg( DWORD type )

Queries the last error message occured.
===============
*/
OGL_ERRORS COGLApplication::DisplayErrorMsg( DWORD type )
{


	return OGL_OK;
}


/*
===============
InitializeOpenGLScene( void )

Sets up basic opengl initalization. Called from CreateOpenGLApplication.
===============
*/
OGL_ERRORS COGLApplication::InitializeOpenGLScene( void )
{
	if( !m_bWindowed )
	{
		DEVMODE ScreenSettings;

		memset( &ScreenSettings, 0, sizeof( DEVMODE ) );

		ScreenSettings.dmSize       = sizeof( DEVMODE );
		ScreenSettings.dmPelsWidth  = m_iWindowWidth;
		ScreenSettings.dmPelsHeight = m_iWindowHeight;
		ScreenSettings.dmBitsPerPel = m_oglSettings.bpp;
		ScreenSettings.dmFields     = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;

		// No throw here												
		if( ChangeDisplaySettings( &ScreenSettings, CDS_FULLSCREEN ) != DISP_CHANGE_SUCCESSFUL )
			MessageBox( m_hWnd, "Could not change display mode.", "ERRPR", MB_OK );
	
	}
	
	m_pfdPixelFormat.nSize           = sizeof( PIXELFORMATDESCRIPTOR );
	m_pfdPixelFormat.nVersion        = 1;
	m_pfdPixelFormat.dwFlags         = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	m_pfdPixelFormat.iPixelType      = PFD_TYPE_RGBA;
	m_pfdPixelFormat.cColorBits      = m_oglSettings.bpp;
	m_pfdPixelFormat.cRedBits        = 0;
	m_pfdPixelFormat.cRedShift       = 0;
	m_pfdPixelFormat.cGreenBits      = 0;
	m_pfdPixelFormat.cGreenShift     = 0;
	m_pfdPixelFormat.cBlueBits       = 0;
	m_pfdPixelFormat.cBlueShift      = 0;
	m_pfdPixelFormat.cAlphaBits      = m_oglSettings.alphabuffer;
	m_pfdPixelFormat.cAlphaShift     = 0;
	m_pfdPixelFormat.cAccumBits      = 0;
	m_pfdPixelFormat.cAccumRedBits   = 0;
	m_pfdPixelFormat.cAccumGreenBits = 0;
	m_pfdPixelFormat.cAccumBlueBits  = 0;
	m_pfdPixelFormat.cAccumAlphaBits = 0;
	m_pfdPixelFormat.cDepthBits      = m_oglSettings.depthbuffer;
	m_pfdPixelFormat.cStencilBits    = m_oglSettings.stencilbuffer;
	m_pfdPixelFormat.cAuxBuffers     = 0;
	m_pfdPixelFormat.iLayerType      = 0;
	m_pfdPixelFormat.bReserved       = 0;
	m_pfdPixelFormat.dwLayerMask     = 0;
	m_pfdPixelFormat.dwVisibleMask   = 0;
	m_pfdPixelFormat.dwDamageMask    = 0;



	return OGL_OK;
}


/*
================
MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )

Windows message procedure function.
================
*/
LRESULT COGLApplication::MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
	switch( msg )
	{
		/*case WM_PAINT:	// Handle paint messages when the app is not ready
			{
				if( m_hDC && !m_bReady )
				{
					if( m_bWindowed )
						SwapBuffers( m_hDC );
				}
            } break;
		*/
		case WM_SIZE:
			{
				ResizeOpenGLScene( LOWORD( lParam ), HIWORD( lParam ) );
				return 0;
			} break;

		case WM_CLOSE:
			{
				CleanUpOpenGLScene();
				DestroyWindow( m_hWnd );
				PostQuitMessage( 0 );

				return 0;
			} break;
	}

	return DefWindowProc( hWnd, msg, wParam, lParam );
}


/*
===============
ToggleFullscreen( void )

Changed from fullscreen->windowed or windowed->fullscreen.
===============
*/
OGL_ERRORS COGLApplication::ToggleFullscreen( void )
{


	return OGL_OK;
}


/*
===============
ForceWindowed( void )

Forces the opengl envirnonment into a window, if the hardware does not 
support opengl in a window, OGL_ERR_WINDOWED_NOT_SUPPORTED returns.
===============
*/
OGL_ERRORS COGLApplication::ForceWindowed( void )
{


	return OGL_OK;
}
	

/*
===============
RenderOpenGLScene( void )

Calls the virtual function Render and also does other rendering tasks like
calculate framerate, check if 3d environment is paused, etc...
===============
*/
OGL_ERRORS COGLApplication::RenderOpenGLScene( void )
{
	OGL_ERRORS err = OGL_OK;

	// Do any resizing code here...

	// Calc the FPS here...

	if( m_bFrameMoving )	// single step would go here also
	{
		err = NextFrame();
		if( err != OGL_OK )
			return err;
	}
	
	err = Render();
	if( err != OGL_OK )
		return err;

	// Keep track of frame count here...

	// Show the newly rendered scene
	if( !SwapBuffers( m_hDC ) )
		return OGL_ERR_CANT_SWAP_BUFFERS;

	return OGL_OK;
}


/*
===============
CleanUpOpenGLScene( void )

Releases all memory when the opengl environment is shutdown.
Call from the MsgProc WM_CLOSE
===============
*/
OGL_ERRORS COGLApplication::CleanUpOpenGLScene( void )
{
	m_bActive = false;
	m_bReady  = false;
	
	CleanupObjects();
	
	if( !m_bWindowed )
		ChangeDisplaySettings( NULL, 0 );

	if( m_hRC )
	{
		if( !wglMakeCurrent( NULL, NULL ) )
			return OGL_ERR_RELEASE_DC_AND_RC_FAILED;

		if( !wglDeleteContext( m_hRC ) )
			return OGL_ERR_RELEASE_RC_FAILED;

		m_hRC = NULL;
	}
	
	if( !ReleaseDC( m_hWnd, m_hDC ) )
	{
		m_hDC = NULL;
		return OGL_ERR_RELEASE_DC_FAILED;
	}

	if( !DestroyWindow( m_hWnd ) )
	{
		m_hWnd = NULL;
		return OGL_ERR_RELEASE_HWND_FAILED;
	}

	ChangeDisplaySettings( 0, 0 );

	return OGL_OK;
}


/*
===============
CreateOpenGLApplication( HINSTANCE hInstance, OGLSettings settings )

Main function to setup opengl. This function will call InitializeOpenGLScene.
This function also sets the fov, number of depthbuffer bits, and bits per fragment (pixel)
===============
*/
OGL_ERRORS COGLApplication::CreateOpenGLApplication( HINSTANCE hInstance, OGLSettings settings )
{
	int pixelFormat = 0;
	OGL_ERRORS error = OGL_OK;
	
	m_oglSettings.bpp           = settings.bpp;
	m_oglSettings.depthbuffer   = settings.depthbuffer;
	m_oglSettings.fov           = settings.fov;
	m_oglSettings.stencilbuffer = settings.stencilbuffer;
	
	m_bWindowed = settings.window;
	if( !m_bWindowed ) m_dwWindowStyle = WS_POPUP;

	if( settings.width != 0 )  m_iWindowWidth  = settings.width;
	if( settings.height != 0 ) m_iWindowHeight = settings.height;

	m_hInstance = hInstance;	

	RECT WindowRect;								// Grabs Rectangle Upper Left / Lower Right Values
	WindowRect.left   = 0L;							// Set Left Value To 0
	WindowRect.right  = (long)m_iWindowWidth;		// Set Right Value To Requested Width
	WindowRect.top    = 0L;							// Set Top Value To 0
	WindowRect.bottom = (long)m_iWindowHeight;		// Set Bottom Value To Requested Height


	if( m_hWnd == NULL )
	{
		WNDCLASS wc;

		//wc.cbSize        = sizeof( WNDCLASSEX );
		wc.hInstance     = hInstance;
		wc.lpfnWndProc   = WndProc;
		wc.lpszClassName = m_strWindowTitle;
		wc.hbrBackground = ( HBRUSH ) GetStockObject( WHITE_BRUSH );
		wc.hCursor       = LoadCursor( NULL, IDC_ARROW );
		wc.hIcon         = LoadIcon( NULL, IDI_APPLICATION );
		//wc.hIconSm       = LoadIcon( NULL, IDI_APPLICATION );
		wc.cbClsExtra    = NULL;
		wc.cbWndExtra    = NULL;
		wc.lpszMenuName  = NULL;
		wc.style         = CS_OWNDC;

		RegisterClass( &wc );

		// Adjust Window To True Requested Size
		AdjustWindowRectEx( &WindowRect, m_dwWindowStyle, NULL, NULL );		

		m_hWnd = CreateWindow( m_strWindowTitle,
							   m_strWindowTitle,
							   m_dwWindowStyle,
							   CW_USEDEFAULT,
							   CW_USEDEFAULT,
							   m_iWindowWidth,
							   m_iWindowHeight,
							   NULL,
							   NULL,
							   hInstance,
							   NULL );
	}	// End if


	// TODO: Do error checking
	InitializeOpenGLScene();


	// Create the contexts
	m_hDC = GetDC( m_hWnd );
	if( !m_hDC ) 
		return OGL_ERR_CREATE_DC_FAILED;

	pixelFormat = ChoosePixelFormat( m_hDC, &m_pfdPixelFormat );
	if( !pixelFormat ) 
		return OGL_ERR_CANT_FIND_PIXELFORMAT;
	
	if( !SetPixelFormat( m_hDC, pixelFormat, &m_pfdPixelFormat ) ) 
		return OGL_ERR_CANT_SET_PIXELFORMAT;

	m_hRC = wglCreateContext( m_hDC );
	if( !m_hRC ) 
		return OGL_ERR_CREATE_RC_FAILED;

	if( !wglMakeCurrent( m_hDC, m_hRC ) ) 
		return OGL_ERR_ACTIVATE_RC_FAILED;


	// Its showtime!
	ShowWindow( m_hWnd, SW_SHOW );
	UpdateWindow( m_hWnd );
	SetForegroundWindow( m_hWnd );
	SetFocus( m_hWnd );


	// Setup the main 3d environment
	error = ResizeOpenGLScene( m_iWindowWidth, m_iWindowHeight );
	if( error != OGL_OK )
		return error;			

	// Initalize one time scene init stuff
	error = InitialSceneInit();
	if( error != OGL_OK )
		return error;
		
	error = InitObjects();
	if( error != OGL_OK )
		return error;
	
	m_bActive = true;	// Set application to active to allow it to run
	m_bReady  = true;

	
	return OGL_OK;
}


/*
===============
SetupForNewDispSettings( void )

DONT USE!!!
===============
*/	
OGL_ERRORS COGLApplication::SetupForNewDispSettings( void )
{
	m_bVidRestart = true;
	m_bActive     = false;

	CreateOpenGLApplication( m_hInstance, m_oglSettings );

	return OGL_OK;
}


/*
===============
RunOpenGLApplication( void )

This function is called by the programmer which calls RenderOpenGLScene.
That function will call the virtual function Render which the programmer will
implement.
===============
*/
int COGLApplication::RunOpenGLApplication( void )
{
	//int iGotMsg;
    MSG  msg;
    
	//PeekMessage( &msg, NULL, 0U, 0U, PM_NOREMOVE );
	while( 1 )//msg.message != WM_QUIT )
    {
        // Use PeekMessage() if the app is active, so we can use idle time to
        // render the scene. Else, use GetMessage() to avoid eating CPU time.
        //if( m_bActive )
        //    iGotMsg = PeekMessage( &msg, NULL, 0, 0, PM_REMOVE );
        //else
          //  iGotMsg = GetMessage( &msg, NULL, 0, 0 );

        if( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) )//iGotMsg )
        {
			if( msg.message == WM_QUIT )
				break;
			else
			{
				TranslateMessage( &msg );
				DispatchMessage( &msg );
			}
        }
        else
        {
			MainGameLoop();

            // Render a frame during idle time (no messages are waiting)
            if( m_bActive && m_bReady )
            {
                if( RenderOpenGLScene() != OGL_OK )
                    SendMessage( m_hWnd, WM_CLOSE, 0, 0 );
            }
			
			// Temp pause routine
			if( KEY_DOWN( VK_F1 ) )
				PauseOpenGLScene( true );
			else if( KEY_DOWN( VK_F2 ) )
				PauseOpenGLScene( false );
			else if( KEY_DOWN( VK_ESCAPE ) )
				SendMessage( m_hWnd, WM_CLOSE, 0, 0 );
        }
    }

    return msg.wParam;
}


/*
===============
PauseOpenGLScene( bool bPause )

Stops all animations.
===============
*/
void COGLApplication::PauseOpenGLScene( bool bPause )
{
	if( bPause )
		m_bActive = false;
	else
		m_bActive = true;

}


/*
===============
SetupVertexProgram( void )


===============
*/
// This should probably be put in initial scene init by the programmer
void COGLApplication::SetupVertexProgram( void )
{
	

}


/*
===============
ChangeResolution( RESOLUTIONS res, WINDOW_MODE winMode )


===============
*/
OGL_ERRORS COGLApplication::ChangeResolution( RESOLUTIONS res, WINDOW_MODE winMode )
{
	m_bReady = false;

	if( winMode == WM_WINDOW )
	{
		m_bWindowed = true;
		m_dwWindowStyle = WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;
	}
	else
	{
		m_bWindowed = false;
		m_dwWindowStyle = WS_POPUP;
	}
	
	switch( res )
	{
		case OGL_RES_320x240:
			{
				m_iOldWindowWidth  = m_iWindowWidth;
				m_iOldWindowHeight = m_iWindowHeight;
				
				m_iWindowWidth  = 320;
				m_iWindowHeight = 240;
			} break;
		
		case OGL_RES_400x300:
			{
				m_iOldWindowWidth  = m_iWindowWidth;
				m_iOldWindowHeight = m_iWindowHeight;
				
				m_iWindowWidth  = 400;
				m_iWindowHeight = 300;
			} break;

		case OGL_RES_512x384:
			{
				m_iOldWindowWidth  = m_iWindowWidth;
				m_iOldWindowHeight = m_iWindowHeight;
				
				m_iWindowWidth  = 512;
				m_iWindowHeight = 384;
			} break;

		case OGL_RES_640x480:
			{
				m_iOldWindowWidth  = m_iWindowWidth;
				m_iOldWindowHeight = m_iWindowHeight;
				
				m_iWindowWidth  = 640;
				m_iWindowHeight = 480;
			} break;

		case OGL_RES_800x600:
			{
				m_iOldWindowWidth  = m_iWindowWidth;
				m_iOldWindowHeight = m_iWindowHeight;
				
				m_iWindowWidth  = 800;
				m_iWindowHeight = 600;
			} break;

		case OGL_RES_1024x768:
			{
				m_iOldWindowWidth  = m_iWindowWidth;
				m_iOldWindowHeight = m_iWindowHeight;
				
				m_iWindowWidth  = 1024;
				m_iWindowHeight = 768;
			} break;

		case OGL_RES_1152x864:
			{
				m_iOldWindowWidth  = m_iWindowWidth;
				m_iOldWindowHeight = m_iWindowHeight;
				
				m_iWindowWidth  = 1152;
				m_iWindowHeight = 864;
			} break;

		case OGL_RES_1280x1024:
			{
				m_iOldWindowWidth  = m_iWindowWidth;
				m_iOldWindowHeight = m_iWindowHeight;
				
				m_iWindowWidth  = 1280;
				m_iWindowHeight = 1024;
			} break;

		case OGL_RES_1600x1200:
			{
				m_iOldWindowWidth  = m_iWindowWidth;
				m_iOldWindowHeight = m_iWindowHeight;
				
				m_iWindowWidth  = 1600;
				m_iWindowHeight = 1200;
			} break;

		case OGL_RES_2048x1536:
			{
				m_iOldWindowWidth  = m_iWindowWidth;
				m_iOldWindowHeight = m_iWindowHeight;
				
				m_iWindowWidth  = 2048;
				m_iWindowHeight = 1536;
			} break;

		case OGL_RES_WS_856x480:
			{
				m_iOldWindowWidth  = m_iWindowWidth;
				m_iOldWindowHeight = m_iWindowHeight;
				
				m_iWindowWidth  = 856;
				m_iWindowHeight = 480;
			} break;

		default: 
			{
				m_iWindowWidth  = m_iOldWindowWidth;
				m_iWindowHeight = m_iOldWindowHeight;

				
				return OGL_ERR_RESOLUTION_NOT_SUPPORTED;
			} break;
	}

	if( !m_bWindowed )
	{
		DEVMODE ScreenSettings;

		memset( &ScreenSettings, 0, sizeof( DEVMODE ) );

		ScreenSettings.dmSize       = sizeof( DEVMODE );
		ScreenSettings.dmPelsWidth  = m_iWindowWidth;
		ScreenSettings.dmPelsHeight = m_iWindowHeight;
		ScreenSettings.dmBitsPerPel = m_oglSettings.bpp;
		ScreenSettings.dmFields     = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;

		// No throw here												
		if( ChangeDisplaySettings( &ScreenSettings, CDS_FULLSCREEN ) != DISP_CHANGE_SUCCESSFUL )
			MessageBox( m_hWnd, "Could not change display mode.", "ERRPR", MB_OK );
	
	}

	CleanUpOpenGLScene();

	if( m_bWindowed )
		m_oglSettings.window = true;
	else
		m_oglSettings.window = false;

	m_oglSettings.width  = m_iWindowWidth;
	m_oglSettings.height = m_iWindowHeight;

	CreateOpenGLApplication( m_hInstance, m_oglSettings );

	m_bReady  = true;
	m_bActive = true;

	return OGL_OK;
}


/*
===============
InitExtentions( const char *ext )


===============
*/
int COGLApplication::InitExtentions( const char *ext )	
{


	return 0;
}


/*
===============
SetTextureShader( const char *shader )


===============
*/
void COGLApplication::SetTextureShader( const char *shader )
{


}


/*
===============
SetRegisterCombinerScript( const char *rc )


===============
*/
void COGLApplication::SetRegisterCombinerScript( const char *rc )
{


}


/*
===============
LoadTextureShaderFromFile( const char *filename, char *texshader )


===============
*/
bool COGLApplication::LoadTextureShaderFromFile( const char *filename )
{


	return true;
}


/*
===============
LoadRegisterCombinerFromFile( const char *filename, char *regcombiner )


===============
*/
bool COGLApplication::LoadRegisterCombinerFromFile( const char *filename )
{


	return true;
}


/*
===============
LoadVertexProgramFromFile( const char *filename, char *vtxprogram )


===============
*/
bool COGLApplication::LoadVertexProgramFromFile( const char *filename )
{
	ifstream fileVtxProg;
	
	int numOfChars   = 0;
	int size         = 0;

	char buffer = 0;
	
	if( !m_bVtxProgReady )
		throw "Vertex Program Not Ready";
	else
	{
		fileVtxProg.open( filename, ios::nocreate );
		if( fileVtxProg.fail() )
		{
			fileVtxProg.close();
			throw "Vertex Program Could Not Open";
		}
		else
		{			
			fileVtxProg.seekg(0, ios::end );
			numOfChars = fileVtxProg.tellg();
			fileVtxProg.seekg(0, ios::beg );			// Go back to the beginning

			// Add one incase a null gets slaped onto the end
			size = (numOfChars / sizeof( char )) + 1;	
					
			m_strVtxProgram = new char[ size ];
						
			// Make sure there is no junk in the array, this may cause a problem
			memset( m_strVtxProgram, '\0', size );
			
			fileVtxProg.read( m_strVtxProgram, size );
		}
	}
	
	fileVtxProg.close();

	return true;
}


/*
===============
SetVertexProgram( const char *shader )


===============
*/
int COGLApplication::SetVertexProgram( GLuint id )
{
	/*if( !m_bVtxProgReady )
		return -1;
	else
	{
		m_pProgID = new GLuint[id];
		
		glGenProgramsNV( 1, m_pProgID );
		glLoadProgramNV( GL_VERTEX_PROGRAM_NV, id, strlen( m_strVtxProgram ), 
			             (const GLubyte*) m_strVtxProgram ); 
		
		if( glGetError() != GL_NO_ERROR )
		{
			return -1;	// Could not load vertex program
		}
	}
*/
	return 0;
}



/*
===============
SetHints( int type, int mode )


===============
*/
void COGLApplication::SetHints( int type, int mode )
{



}
	
