/////////////////////////////////////////////
// Infinite Stencil Shadow Volumes
// -------------------------------
// By: Thomas Grimes
// Date: 9-19-02
/////////////////////////////////////////////

#define GLH_EXT_SINGLE_FILE	1
// Some of these files already include the other glh headers
#include <nvgl\glh_extensions.h>
#include <nvgl\glh_array.h>
#include <nvgl\glh_combiners.h>
//#include <nvgl\glh_cube_map.h>
#include <nvgl\glh_mipmaps.h>
///#include <nvgl\glh_text.h>


#include <string.h>
#include <math.h>

#include "OpenGL Application.h"
#include "matrix.h"


#define NEAR_EPSILON 0.01f
#define NEAR_PLANE   1.0f
#define FAR_PLANE	 1000.0f

#define SHOW_VOLUME 0

//
/// Light stuff
//

GLfloat ambient[]  = { 0.15f,  0.15f, 0.15f, 1.0f };
GLfloat diffuse[]  = { 1.0f,  1.0f, 1.0f, 1.0f };
GLfloat specular[] = { 1.0f,  1.0f, 1.0f, 1.0f };
GLfloat lpos[]     = { 0.0f, 5.0f, 0.0f, 1.0f };


// 
/// Occluder Stuff
//
float cube_verts[] =
{
	 // Back
	 2.0f, 2.0f, -2.0f, 1.0f,
	 2.0f, -2.0f, -2.0f, 1.0f,
	-2.0f, -2.0f, -2.0f, 1.0f,
	-2.0f, 2.0f, -2.0f, 1.0f,
		
	 // Right
	 2.0f, 2.0f, 2.0f, 1.0f,
	 2.0f, -2.0f, 2.0f, 1.0f,
	 2.0f, -2.0f, -2.0f, 1.0f,
	 2.0f, 2.0f, -2.0f, 1.0f,	 

	 // Left
	-2.0f, 2.0f, -2.0f, 1.0f,
	-2.0f, -2.0f, -2.0f, 1.0f,
	-2.0f, -2.0f, 2.0f, 1.0f,
    -2.0f, 2.0f, 2.0f, 1.0f,    

	 // Front
	 2.0f, 2.0f, 2.0f, 1.0f,
    -2.0f, 2.0f, 2.0f, 1.0f,
    -2.0f, -2.0f, 2.0f, 1.0f,
	 2.0f, -2.0f, 2.0f, 1.0f,

	 // Top
	 2.0f, 2.0f, -2.0f, 1.0f,
    -2.0f, 2.0f, -2.0f, 1.0f,
    -2.0f, 2.0f, 2.0f, 1.0f,
	 2.0f, 2.0f, 2.0f, 1.0f,

	 // Bottom
	 2.0f, -2.0f, -2.0f, 1.0f,
	 2.0f, -2.0f, 2.0f, 1.0f,
	-2.0f, -2.0f, 2.0f, 1.0f,
    -2.0f, -2.0f, -2.0f, 1.0f	 
};

float cube_normals[] =
{
	0.0f, 0.0f, -1.0f,
	1.0f, 0.0f, 0.0f, 
	-1.0f, 0.0f, 0.0f, 
	0.0f, 0.0f, 1.0f, 
	0.0f, 1.0f, 0.0f, 
	0.0f, -1.0f, 0.0f
};

struct vert
{
	float x, y, z, w;
};

struct plane
{
	float nx, ny, nz;
	float d;
};

struct trilist
{
	vert  polyVerts[4];
	plane polyPlane;
	bool  backFacing;
	int   adjacent[4];
};

trilist triList[6];

inline float DP3( vert &v1, vert &v2 )
{
	return ( v1.x * v2.x + v1.y * v2.y + v1.z * v2.z );
}

// CrossProduct
inline vert XPD( vert &v1, vert &v2 )
{
	vert v;

	v.x = ( v1.y * v2.z ) - ( v1.z * v2.y );
	v.y = ( v1.z * v2.x ) - ( v1.x * v2.z );
	v.z = ( v1.x * v2.y ) - ( v1.y * v2.x );

	return v;
}

inline float PolyFacingWRTLight( vert &normal, float polyD,  vert &light )
{
	// Solve plane equation of poly with respect to the light to determine facingness
	// For d < 0  : triangle is backfacing
	// For d >= 0 : tirangle is frontfacing
	float d = DP3( normal, light ) + polyD * light.w;

	return d;
}

void SetupOccluder( void )
{
	int v = 0;
	int x = 0, y = 0;

	float len = 0.0f;

	for( x = 0; x < 6; ++x )
	{
		for( y = 0; y < 4; ++y )
		{
			triList[x].polyVerts[y].x = cube_verts[v++];
			triList[x].polyVerts[y].y = cube_verts[v++];
			triList[x].polyVerts[y].z = cube_verts[v++];
			triList[x].polyVerts[y].w = cube_verts[v++];
		}
	}		

	// Compute plane equation for each polygon
	vert norm;//, v1, v2;
	float d = 0.0f;
	v = 0;
	for( x = 0; x < 6; ++x )
	{
		/*v1.x = triList[x].polyVerts[1].x - triList[x].polyVerts[0].x;   
		v1.y = triList[x].polyVerts[1].y - triList[x].polyVerts[0].y;
		v1.z = triList[x].polyVerts[1].z - triList[x].polyVerts[0].z;

		v2.x = triList[x].polyVerts[3].x - triList[x].polyVerts[0].x;   
		v2.y = triList[x].polyVerts[3].y - triList[x].polyVerts[0].y;
		v2.z = triList[x].polyVerts[3].z - triList[x].polyVerts[0].z;

		norm = XPD( v1, v2 );
		d    = -DP3( norm, triList[x].polyVerts[0] ); 

		len = (float)sqrt( pow( norm.x, 2 ) + pow( norm.y, 2 ) + pow( norm.z, 2 ) + pow( norm.w, 2 ) );		
		
		if( len == 0 )
		{
			triList[x].polyPlane.nx = 0.0f;
			triList[x].polyPlane.ny = 0.0f;
			triList[x].polyPlane.nz = 0.0f;
		}

		norm.x *= len;
		norm.y *= len;
		norm.z *= len;
		norm.w *= len;
		*/

		norm.x = cube_normals[v++];
		norm.y = cube_normals[v++];
		norm.z = cube_normals[v++];

		d = -DP3( norm, triList[x].polyVerts[0] ); 

		triList[x].polyPlane.nx = norm.x;
		triList[x].polyPlane.ny = norm.y;
		triList[x].polyPlane.nz = norm.z;
		triList[x].polyPlane.d  = d;		
	}
}

void DetermineFacingness( void )
{
	vert n;
	float d;
	float face;
	vert l;
	
	l.x = lpos[0]; l.y = lpos[1]; l.z = lpos[2]; l.w = lpos[3];
	for( int x = 0; x < 6; ++x  )
	{
		n.x = triList[x].polyPlane.nx;
		n.y = triList[x].polyPlane.ny;
		n.z = triList[x].polyPlane.nz;
		d   = triList[x].polyPlane.d;

		face = PolyFacingWRTLight( n, d, l );

		// For d < 0  : triangle is backfacing
		// For d >= 0 : tirangle is frontfacing
		if( face < EPSILON )
			triList[x].backFacing = true;
		else
			triList[x].backFacing = false;
	}
}

struct edge_t
{
	int v0, v1;
};

bool CompareEdges( const unsigned int p1, const unsigned int p2, const edge_t e1, const edge_t e2 )
{
	// Edge compare should make sure to test if edge pair is flipped
	// Sometimes the edges are the same but one is flipped, ex:
	// Edge compare should be checked like so:
	//    Where bn & rn are indices to vertices
	//		Let edgeSet1 = {b1->b2}
	//		Let edgeSet2 = {r1->r2}
	//          if( b1->b2 == r1->r2 OR b1->b2 == r2->r1 ) matchFound = true;

	bool edgeTest[3];

	edgeTest[0] = ( triList[p1].polyVerts[e1.v0].x == triList[p2].polyVerts[e2.v0].x &&
		            triList[p1].polyVerts[e1.v1].x == triList[p2].polyVerts[e2.v1].x ) || 
				  ( triList[p1].polyVerts[e1.v0].x == triList[p2].polyVerts[e2.v1].x &&
				    triList[p1].polyVerts[e1.v1].x == triList[p2].polyVerts[e2.v0].x );
	
	edgeTest[1] = ( triList[p1].polyVerts[e1.v0].y == triList[p2].polyVerts[e2.v0].y &&
		            triList[p1].polyVerts[e1.v1].y == triList[p2].polyVerts[e2.v1].y ) || 
				  ( triList[p1].polyVerts[e1.v0].y == triList[p2].polyVerts[e2.v1].y &&
				    triList[p1].polyVerts[e1.v1].y == triList[p2].polyVerts[e2.v0].y );

	edgeTest[2] = ( triList[p1].polyVerts[e1.v0].z == triList[p2].polyVerts[e2.v0].z &&
		            triList[p1].polyVerts[e1.v1].z == triList[p2].polyVerts[e2.v1].z ) || 
				  ( triList[p1].polyVerts[e1.v0].z == triList[p2].polyVerts[e2.v1].z &&
				    triList[p1].polyVerts[e1.v1].z == triList[p2].polyVerts[e2.v0].z );

	if( edgeTest[0] && edgeTest[1] && edgeTest[2] )
		return true;	
	
	return false;
}

/*  Adjacency Algorithm
	-------------------
 
    ForEach polygon in TriList
		Setup edge list
	
	ForEach polygon in TriList (loop var x)
		ForEach edge in EdgeList
			ForEach polygon in TriList (loop var y)
				If edge is similar to anyother edge in other polygons
					Let triList[x].adjacent[adj++] = y
					Break
				EndIf
			Next y
		Next edge
	Next x
	
*/
void SetupAdjacency( void )
{
	int x, y;
	int adj = 0;

	bool foundAdj = false;

	edge_t edges[6][4]; // 6 polygons with 4 edges per polygon
	edge_t e, ee;

	for( x = 0; x < 6; ++x )
		for( y = 0; y < 4; ++y )
		{
			edges[x][y].v0 = (y + 0) % 4;
			edges[x][y].v1 = (y + 1) % 4;
		}
	
	for( x = 0; x < 6; ++x )	// Go through each polygon
	{
		adj = 0;		
		for( int edge = 0; edge < 4; ++edge )   // Go through this polygon's edges
		{
			e.v0 = edges[x][edge].v0;	// Save current edge
			e.v1 = edges[x][edge].v1;
			for( y = 0; y < 6; ++y )	// Go through the polygons to test agianst
			{
				foundAdj = false;
				if( x != y )	// If the two polygons are not the same
					for( int edge2 = 0; edge2 < 4; ++edge2 )	// Compare the two edges
					{
						ee.v0 = edges[x][edge2].v0;
						ee.v1 = edges[x][edge2].v1;
						if( CompareEdges( x, y, e, ee ) )
						{
							triList[x].adjacent[adj++] = y;
							foundAdj = true;
						}
						
						if( foundAdj )
							break;
					}
			}
		}
	}
}


// Converts a normal [-1,1] to an rgb value of [0,255]
// -1 <= n.(x,y,z) <= 1  to:
//		-1*127+127 <= n.(x,y,z) <= 1*127+127
//				Evaluates to: (with rounding errors)
//		 0 <= n.(x,y,z) <= 255
void vec_to_rgb( VECTOR &v, unsigned char *ip )
{
	*(ip++) = (unsigned char) ((v.x * 127.5f) + 127.5f);
	*(ip++) = (unsigned char) ((v.y * 127.5f) + 127.5f);
	*(ip++) = (unsigned char) ((v.z * 127.5f) + 127.5f);
}

const GLuint g_cubeMapDirections[6] = 
{
	GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB,

	GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB,

	GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB
};


CMyOpenGLApp::CMyOpenGLApp() 
{
	strcpy( m_strWindowTitle, "OpenGL Application" );

	// TODO: Add any initialization code here...
}


OGL_ERRORS CMyOpenGLApp::InitialSceneInit( void )
{
	char buffer[1024] = {0};
	char *tbuffer = new char[1024];

	bool noext = false;

	glClearColor( 1.0f, 1.0f, 1.0f, 1.0f );
	
	glEnable( GL_DEPTH_TEST );
	glDisable( GL_LIGHTING );
	
	glShadeModel( GL_SMOOTH );
	
		
	
	/*
	if( !glh_init_extensions( "GL_NV_vertex_program "
							  "GL_VERSION_1_2 "
							  "GL_ARB_multitexture "
							  "GL_EXT_secondary_color "
							  "GL_NV_register_combiners "
							  "GL_ARB_texture_cube_map "
							) )
	{
		noext = true;
	}
	*/

	if( !glh_init_extensions( "GL_VERSION_1_2 "
							  "GL_ARB_texture_cube_map "
							) )
	{
		noext = true;
	}


	if( noext )
	{
		memset( tbuffer, 0, sizeof( char ) * 1024 );		

		strcpy( buffer, "Could not initialize these extensions:\n"
			            "-------------------------------------------\n\n" );
		strcpy( tbuffer, glh_get_unsupported_extensions() );
	
		int len = strlen( tbuffer );
		for( int x = 0; x < len; ++x )
		{
			if( tbuffer[x] == ' ' )
				tbuffer[x] = '\n';
		}

		strcat( buffer, tbuffer );

		delete [] tbuffer;
		tbuffer = 0;
		
		throw buffer;
	}

	m_bVtxProgReady = true;

	
	// TODO: Setup any texture loading here... 
	glGenTextures( 1, &wall );
	glBindTexture( GL_TEXTURE_2D, wall );

	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
	
	// Make a nice checkerboard texture
	GLfloat *tex = new GLfloat[32*32];
	{
		for(int i=0; i < 32; i++)
		{
			for(int j=0; j < 32; j++)
			{
				if(i>>4 ^ j>>4)
					tex[i+j*32] = 1.0f;
				else
					tex[i+j*32] = 0.9f;
			}
		}
	}

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 32, 32, 0, GL_LUMINANCE, GL_FLOAT, tex);
	delete [] tex;

	SetupOccluder();


	return OGL_OK;
}

OGL_ERRORS CMyOpenGLApp::InitObjects( void )                        
{
	// Load general transform program
	//glGenProgramsNV( 1, &m_vtxprog_gen );
	//glLoadProgramNV( GL_VERTEX_PROGRAM_NV, m_vtxprog_gen, strlen( vtxprog_gen ), (const GLubyte*) vtxprog_gen );
	//if( glGetError() != GL_NO_ERROR )
	//	throw "Could not load 'general' Vertex Program (PPL)";
	
	// Setup list to pass the texture right on through into the output
	// Used if no fancy combiner math ops are needed
	rclist_passthrough = glGenLists(1);
	glNewList( rclist_passthrough, GL_COMPILE );

		nvparse(
		"!!RC1.0											\n"
		"out.rgb = tex0;				     				\n"
		"out.a   = unsigned_invert( zero );					\n"
		);
#ifdef _DEBUG	
	{
		// This code is bugged!
		bool flag = false;
		for( char *const *errors = nvparse_get_errors(); *errors; ++errors )
		{
			flag = true;

			MessageBox( NULL, (char*)errors, "Error in nvparse", MB_OK );
		}
		if( flag ) throw "Error in nvparse.\nProgram will now close.";
	}
#endif

	glEndList();


	// TODO: Setup objects such as vertex programs, register combiners, display lists, etc...

	// Setup occluder model adjacency information
	SetupAdjacency();

	return OGL_OK;
}

OGL_ERRORS CMyOpenGLApp::RestoreObjects( void )   
{
	// TODO: Reload any objects which get destroyed if the window size/mode changes...
	
	return OGL_OK;
}

OGL_ERRORS CMyOpenGLApp::ResizeOpenGLScene( int width, int height ) 
{
	float matrix[4][4] = {0};
	
	glViewport( 0, 0, width, height );
	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	
	gluPerspective( m_oglSettings.fov, (GLfloat) width / (GLfloat) height, NEAR_PLANE, FAR_PLANE );

	glGetFloatv( GL_PROJECTION_MATRIX, &matrix[0][0] );

	// Nudge infinity in just slightly for lsb slop
	// This can fix problems some cards have with the 'razor's edge' at infinity
    //float nudge = 1 - 1.0 / (1<<23);
	float nudge = 0.99999988f;
	
	// Set far plane to infinity
	matrix[3][2] = -2 * NEAR_PLANE * nudge;
	matrix[2][2] = -1 * nudge;

	glLoadMatrixf( &matrix[0][0] );	
	
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	
	return OGL_OK;
}

OGL_ERRORS CMyOpenGLApp::NextFrame( void )	
{
	static float theta  = 0.0f;

	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();

	
	// TODO: Add any frame updating code here...
	lpos[0] = cos( theta ) * 10.0f;
	lpos[2] = sin( theta ) * 10.0f;

	
	theta += 0.005f;
	if( theta >= 2*3.1415 )
		theta = 0.0f;

	return OGL_OK;
}

OGL_ERRORS CMyOpenGLApp::Render( void )   
{
	glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
	
	static float zero[] = { 0.0f, 0.0f, 0.0f, 0.0f };

	int x, y;	// For the loops
	

	glTranslatef( 0.0f, 0.0f, -20.0f );
	glRotatef( 20.0f, 1.0f, 1.0f, 0.0f );
		
	glEnable( GL_TEXTURE_2D );
	glEnable( GL_DEPTH_TEST );
	glEnable( GL_CULL_FACE );
					
	glCullFace( GL_BACK );
	glFrontFace( GL_CCW );
		
#if !SHOW_VOLUME

	// Draw base pass to fill the depth buffer with initial info
	glEnable( GL_STENCIL_TEST );

	glStencilFunc(GL_ALWAYS, 0, ~0);
	glStencilMask( ~0 );
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

	glDepthFunc( GL_LESS );

	DrawRoom();
	DrawBox();
				
	// Now the shadow passes
	glDepthMask( 0 );
	glColorMask( 0, 0, 0, 0 );
		
	// Now, configure zfail to increment
	glCullFace( GL_FRONT );
	glStencilOp( GL_KEEP, GL_INCR, GL_KEEP );

#endif		
	
	//CMatrix matrix;
	//vector_t vecL, oVecL;
	float L[4];
	//double modelview[16];
	//double invmodelview[16];

	L[0] = lpos[0]; L[1] = lpos[1]; L[2] = lpos[2]; L[3] = lpos[3];
	//vecL.x = lpos[0]; vecL.y = lpos[1]; vecL.z = lpos[2]; vecL.w = lpos[3];

	//glGetDoublev( GL_MODELVIEW_MATRIX, modelview );
	
	//matrix.StoreDouble16( modelview );

	//invertMatrixf( (float*)invmodelview, (float*)modelview );

	//matrix.StoreDouble16( invmodelview );
	//matrix.Inverse( matrix );
	//matrix.Multiply( vecL, oVecL );

	//L[0] = oVecL.x; L[1] = oVecL.y; L[2] = oVecL.z; L[3] = oVecL.w;
		
#if !SHOW_VOLUME
		
	// Determine whether each triangle in the occluder's model is front- or back-facing
	// with respect to the light's position 
	DetermineFacingness();

	// Now render all possible silhouette edges as quads projected away from the light
	// to infinity
	glBegin( GL_QUADS );
	for( x = 0; x < 6; ++x )			 // For each polygon
		if( !triList[x].backFacing )	 // If polygon _is_ front facing
			for( y = 0; y < 4; ++y )	 // For each polygon edge
				if( triList[triList[x].adjacent[y]].backFacing ) // If adjacent _is_ backfacing 
				{
					// We found a possible sil edge
					vert A = triList[x].polyVerts[y];
					vert B = triList[x].polyVerts[(y+1)%4];

					glVertex4f( B.x, B.y, B.z, B.w );
					glVertex4f( A.x, A.y, A.z, A.w );

					// Infinite verts
					glVertex4f( A.x*L[3] - L[0]*A.w,
						        A.y*L[3] - L[1]*A.w,
								A.z*L[3] - L[2]*A.w, 0 );
					glVertex4f( B.x*L[3] - L[0]*B.w,
						        B.y*L[3] - L[1]*B.w,
								B.z*L[3] - L[2]*B.w, 0 );
				}
	glEnd();

#define V triList[x].polyVerts[y] // macro used in Vertex4f calls

	// Now render the volume caps
	glBegin( GL_QUADS );
		for( x = 0; x < 6; ++x )
			if( triList[x].backFacing )
				for( y = 0; y < 4; ++y )
					glVertex4f( V.x*L[3] - L[0]*V.w, 
					            V.y*L[3] - L[1]*V.w,
								V.z*L[3] - L[2]*V.w, 0 ); // Infinite
			else
				for( y = 0; y < 4; ++y )
					glVertex4f( V.x, V.y, V.z, V.w );
	glEnd();
		
		
	//
	/// Now configure zfail to decrement for front facing polygons
	//
	glCullFace( GL_BACK );
	glStencilOp( GL_KEEP, GL_DECR, GL_KEEP );

	// Now render the volume again but with front facing polygons rather than back facing
	glBegin( GL_QUADS );
	for( x = 0; x < 6; ++x )			 // For each polygon
		if( !triList[x].backFacing )	 // If polygon _is_ front facing
			for( y = 0; y < 4; ++y )	 // For each polygon edge
				if( triList[triList[x].adjacent[y]].backFacing ) // If adjacent _is_ backfacing 
				{
					// We found a possible sil edge
					vert A = triList[x].polyVerts[y];
					vert B = triList[x].polyVerts[(y+1)%4];
						
					glVertex4f( B.x, B.y, B.z, B.w );
					glVertex4f( A.x, A.y, A.z, A.w );

					// Infinite verts
					glVertex4f( A.x*L[3] - L[0]*A.w,
						        A.y*L[3] - L[1]*A.w,
								A.z*L[3] - L[2]*A.w, 0 );
					glVertex4f( B.x*L[3] - L[0]*B.w,
						        B.y*L[3] - L[1]*B.w,
								B.z*L[3] - L[2]*B.w, 0 );
				}
	glEnd();

	// Now render the volume caps
	glBegin( GL_QUADS );
		for( x = 0; x < 6; ++x )
			if( triList[x].backFacing )
				for( y = 0; y < 4; ++y )
					glVertex4f( V.x*L[3] - L[0]*V.w, 
					            V.y*L[3] - L[1]*V.w,
								V.z*L[3] - L[2]*V.w, 0 ); // Infinite
			else
				for( y = 0; y < 4; ++y )
					glVertex4f( V.x, V.y, V.z, V.w );
	glEnd();

#endif
	// Paint the light
		
	glEnable( GL_LIGHTING );
	glEnable( GL_LIGHT0 );

	glLightfv( GL_LIGHT0, GL_AMBIENT, ambient );
	glLightfv( GL_LIGHT0, GL_DIFFUSE, diffuse );
	glLightfv( GL_LIGHT0, GL_SPECULAR, specular );
	glLightfv( GL_LIGHT0, GL_POSITION, lpos );

	//glDisable( GL_CULL_FACE );

#if !SHOW_VOLUME
		
	// Setup stencil to only put light where there is no shadow (where stencil is 0)
	glStencilFunc( GL_EQUAL, 0, ~0 );
	glDepthFunc( GL_EQUAL );
		
	glStencilOp( GL_KEEP, GL_KEEP, GL_KEEP );
	
	glColorMask( 1, 1, 1, 1 );
	glDepthMask( 1 );

	glEnable( GL_BLEND );
	glBlendFunc( GL_ONE, GL_ONE );

#endif

	DrawRoom();
	DrawBox();
		

#if SHOW_VOLUME
		//////////////////////////
#define V triList[x].polyVerts[y] // macro used in Vertex4f calls
		
	// These two uncommented make the volume translucent
	glEnable( GL_BLEND );
	glBlendFunc( GL_ONE, GL_ONE );

	DetermineFacingness();
	
	glDepthFunc(GL_LESS);
	glDepthMask(0);

	//glEnable(GL_BLEND);
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	//glColor4f(1,0,0,.7 * .1);
	//draw_shadow_volume_end_caps(mindex);
	// Now render the volume caps
	//glRotatef( 20.0f, 0.0f, 1.0f, 0.0f );

	glBegin( GL_QUADS );
		for( x = 0; x < 6; ++x )
			if( triList[x].backFacing )
				for( y = 3; y >= 0; --y )
					glVertex4f( V.x*L[3] - L[0]*V.w, 
					            V.y*L[3] - L[1]*V.w,
								V.z*L[3] - L[2]*V.w, 0 ); // Infinite
			else
				for( y = 0; y < 4; ++y )
					glVertex4f( V.x, V.y, V.z, V.w );
	glEnd();

		
	//glColor4f(1,0,.7,.15 * .1);
	//draw_shadow_volume_extruded_edges(mindex);
	// Now render all possible silhouette edges as quads projected away from the light
	// to infinity
	glBegin( GL_QUADS );
	for( x = 0; x < 6; ++x )		 // For each polygon
		if( !triList[x].backFacing )	 // If polygon _is_ front facing
			for( y = 0; y < 4; ++y ) // For each polygon edge
				if( triList[triList[x].adjacent[y]].backFacing ) // If adjacent _is_ backfacing 
				{
					// We found a possible sil edge
					vert A = triList[x].polyVerts[y];
					vert B = triList[x].polyVerts[(y+1)%4];

					glVertex4f( B.x, B.y, B.z, B.w );
					glVertex4f( A.x, A.y, A.z, A.w );

					// Infinite verts
					glVertex4f( A.x*L[3] - L[0]*A.w,
						        A.y*L[3] - L[1]*A.w,
								A.z*L[3] - L[2]*A.w, 0 );
					glVertex4f( B.x*L[3] - L[0]*B.w,
						        B.y*L[3] - L[1]*B.w,
								B.z*L[3] - L[2]*B.w, 0 );
				}
	glEnd();

	glDepthMask(1);
	glDepthFunc(GL_LESS);
	glDisable(GL_BLEND);
	
	//////////////////////////
#endif
	
	glDepthFunc( GL_LESS );		

	glDisable( GL_BLEND );
	glDisable( GL_STENCIL_TEST );
	glDisable( GL_LIGHT0 );
	glDisable( GL_LIGHTING );
	glDisable( GL_TEXTURE_2D );

	// Draw the light source
	glPushMatrix();
	glPointSize( 5.0f );
	glBegin( GL_POINTS );
	
		glColor3f( 1.0f, 0.0f, 0.0f );
		glVertex3f( lpos[0], lpos[1], lpos[2] );

	glEnd();
	glPopMatrix();

	// Set the color to black b/c if not, the last color used will be blened with the scene
	// which will/may give unwanted results
	glColor3f( 0.0f, 0.0f, 0.0f );
	

	return OGL_OK;
}

OGL_ERRORS CMyOpenGLApp::CleanupObjects( void )
{
	// TODO: Destroy any texture objects, display lists, etc here...

	return OGL_OK;
}


/* Cube map layout

			-------
		   |       |
		   |   +y  |
		   |       |
		   |       |
	--------------------------------
   |       |       |       |        |
   |       |       |       |        |
   |   -x  |   +z  |   +x  |    -z  |
   |       |       |       |        |
    --------------------------------
	       |       |
		   |       |
		   |   -y  |
		   |       |
		    -------

*/
// Code taken from NVIDIA cube map header, modified by me
// According to NVIDIA, size of 32 is often sufficient
GLuint CMyOpenGLApp::SetupNormalCubeMap( const int size )
{
	// Allocates for the area of each side of the cube map (width*height) and multiplies
	// by three because each pixel is three bytes wide (24bit RGB)
	unsigned char *data = new unsigned char [size * size * 3];
	unsigned char *ip;
	
	const int level = 0;

	float offset   = 0.5f;
	float delta    = 1;
	float halfsize = size / 2.0f;
	
	VECTOR v;	// Computed normal for each pixel in the cubemap

	GLuint texture;

	// Setup the normalization cube map
	glEnable( GL_TEXTURE_CUBE_MAP_ARB );

	glGenTextures( 1, &texture );

	glBindTexture( GL_TEXTURE_CUBE_MAP_ARB, texture );

	glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

	glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
	glTexParameteri( GL_TEXTURE_CUBE_MAP_ARB, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE );
	
	// Make sure enough memeory can be allocated before trying to use it
	if( !data ) throw "Could not allocate enough memory for the NormalCubeMap";
	

	// positive x image	
	{
		ip = data;
		for(int j = 0; j < size; j++)
		{
			for(int i=0; i < size; i++)
			{
				v.z = -(i*delta + offset - halfsize);
				v.y = -(j*delta + offset - halfsize);
				v.x = halfsize;
				Normalize( v );
				vec_to_rgb(v, ip);				
				ip += 3;
			}
		}
		glTexImage2D( g_cubeMapDirections[0],
					  level, GL_RGB8, size, size, 0, GL_RGB, GL_UNSIGNED_BYTE, data );
	}

	// negative x image	
	{
		ip = data;
		for(int j = 0; j < size; j++)
		{
			for(int i=0; i < size; i++)
			{
				v.z = (i*delta + offset - halfsize);
				v.y = -(j*delta + offset - halfsize);
				v.x = -halfsize;
				Normalize( v );
				vec_to_rgb(v, ip);				
				ip += 3;
			}
		}
		glTexImage2D( g_cubeMapDirections[1],
					  level, GL_RGB8, size, size, 0, GL_RGB, GL_UNSIGNED_BYTE, data );
	}
	
	// positive y image	
	{
		ip = data;
		for(int j = 0; j < size; j++)
		{
			for(int i=0; i < size; i++)
			{
				v.x = (i*delta + offset - halfsize);
				v.z = (j*delta + offset - halfsize);
				v.y = halfsize;
				Normalize( v );
				vec_to_rgb(v, ip);				
				ip += 3;
			}
		}
		glTexImage2D( g_cubeMapDirections[2],
					  level, GL_RGB8, size, size, 0, GL_RGB, GL_UNSIGNED_BYTE, data );
	}

	// negative. y image	
	{
		ip = data;
		for(int j = 0; j < size; j++)
		{
			for(int i=0; i < size; i++)
			{
				v.x = (i*delta + offset - halfsize);
				v.z = -(j*delta + offset - halfsize);
				v.y = -halfsize;
				Normalize( v );
				vec_to_rgb(v, ip);				
				ip += 3;
			}
		}
		glTexImage2D( g_cubeMapDirections[3],
					  level, GL_RGB8, size, size, 0, GL_RGB, GL_UNSIGNED_BYTE, data );
	}
	
	// positive z image	
	{
		ip = data;
		for(int j = 0; j < size; j++)
		{
			for(int i=0; i < size; i++)
			{
				v.x = (i*delta + offset - halfsize);
				v.y = -(j*delta + offset - halfsize);
				v.z = halfsize;
				Normalize( v );
				vec_to_rgb(v, ip);				
				ip += 3;
			}
		}
		glTexImage2D( g_cubeMapDirections[4],
					  level, GL_RGB8, size, size, 0, GL_RGB, GL_UNSIGNED_BYTE, data );
	}

	// negative z image	
	{
		ip = data;
		for(int j = 0; j < size; j++)
		{
			for(int i=0; i < size; i++)
			{
				v.x = -(i*delta + offset - halfsize);
				v.y = -(j*delta + offset - halfsize);
				v.z = -halfsize;
				Normalize( v );
				vec_to_rgb(v, ip);				
				ip += 3;
			}
		}
		glTexImage2D( g_cubeMapDirections[5],
					  level, GL_RGB8, size, size, 0, GL_RGB, GL_UNSIGNED_BYTE, data );
	}

	delete data;

	return texture;	
}

void CMyOpenGLApp::DrawRoom( void )
{
	static float x = 20.0f;
	static float y = 5.0f;

	glBindTexture( GL_TEXTURE_2D, wall );

	glPushMatrix();

	//glRotatef( 20.0f, 0.0f, 1.0f, 0.0f );

	// Back wall
	glBegin( GL_QUADS );

		glTexCoord2f( y, y );
		glNormal3f( 0.0f, 0.0f, 1.0f );
		glVertex3f( x, x, -x );
		
		glTexCoord2f( 0.0f, y );
		glNormal3f( 0.0f, 0.0f, 1.0f );
		glVertex3f( -x, x, -x );

		glTexCoord2f( 0.0f, 0.0f );
		glNormal3f( 0.0f, 0.0f, 1.0f );
		glVertex3f( -x, -x, -x );

		glTexCoord2f( y, 0.0f );
		glNormal3f( 0.0f, 0.0f, 1.0f );
		glVertex3f( x, -x, -x );

	glEnd();

	// Right wall
	glBegin( GL_QUADS );

		glTexCoord2f( y, y );
		glNormal3f( -1.0f, 0.0f, 0.0f );
		glVertex3f( x, x, x );

		glTexCoord2f( 0.0f, y );
		glNormal3f( -1.0f, 0.0f, 0.0f );
		glVertex3f( x, x, -x );

		glTexCoord2f( 0.0f, 0.0f );
		glNormal3f( -1.0f, 0.0f, 0.0f );
		glVertex3f( x, -x, -x );

		glTexCoord2f( y, 0.0f );
		glNormal3f( -1.0f, 0.0f, 0.0f );
		glVertex3f( x, -x, x );

	glEnd();

	// Left wall
	glBegin( GL_QUADS );

		glTexCoord2f( y, y );
		glNormal3f( 1.0f, 0.0f, 0.0f );
		glVertex3f( -x, x, -x );

		glTexCoord2f( 0.0f, y );
		glNormal3f( 1.0f, 0.0f, 0.0f );
		glVertex3f( -x, x, x );

		glTexCoord2f( 0.0f, 0.0f );
		glNormal3f( 1.0f, 0.0f, 0.0f );
		glVertex3f( -x, -x, x );

		glTexCoord2f( y, 0.0f );
		glNormal3f( 1.0f, 0.0f, 0.0f );
		glVertex3f( -x, -x, -x );

	glEnd();

	// Front wall
	glBegin( GL_QUADS );

		glTexCoord2f( y, y );
		glNormal3f( 0.0f, 0.0f, -1.0f );
		glVertex3f( x, x, x );

		glTexCoord2f( y, 0.0f );
		glNormal3f( 0.0f, 0.0f, -1.0f );
		glVertex3f( x, -x, x );

		glTexCoord2f( 0.0f, 0.0f );
		glNormal3f( 0.0f, 0.0f, -1.0f );
		glVertex3f( -x, -x, x );

		glTexCoord2f( 0.0f, y );
		glNormal3f( 0.0f, 0.0f, -1.0f );
		glVertex3f( -x, x, x );		

	glEnd();

	// Ceiling
	glBegin( GL_QUADS );

		glTexCoord2f( y, y );
		glNormal3f( 0.0f, -1.0f, 0.0f );
		glVertex3f( x, x, -x );

		glTexCoord2f( y, 0.0f );
		glNormal3f( 0.0f, -1.0f, 0.0f );
		glVertex3f( x, x, x );

		glTexCoord2f( 0.0f, 0.0f );
		glNormal3f( 0.0f, -1.0f, 0.0f );
		glVertex3f( -x, x, x );

		glTexCoord2f( 0.0f, y );
		glNormal3f( 0.0f, -1.0f, 0.0f );
		glVertex3f( -x, x, -x );	

	glEnd();

	// Floor
	glBegin( GL_QUADS );

		glTexCoord2f( y, y );
		glNormal3f( 0.0f, 1.0f, 0.0f );
		glVertex3f( x, -x, -x );

		glTexCoord2f( 0.0f, y );
		glNormal3f( 0.0f, 1.0f, 0.0f );
		glVertex3f( -x, -x, -x );

		glTexCoord2f( 0.0f, 0.0f );
		glNormal3f( 0.0f, 1.0f, 0.0f );
		glVertex3f( -x, -x, x );

		glTexCoord2f( y, 0.0f );
		glNormal3f( 0.0f, 1.0f, 0.0f );
		glVertex3f( x, -x, x );

	glEnd();

	glPopMatrix();
}

void CMyOpenGLApp::DrawBox( void )
{
	//static float x = 2.0f;

	glPushMatrix();

	//glRotatef( 20, 0.0f, 1.0f, 0.0f );

	//glColor3f( 0.0f, 1.0f, 0.0f );

	for( int x = 0; x < 6; ++x )
	{
		glBegin( GL_QUADS );
		for( int y = 0; y < 4; ++y )
		{
			glNormal3f( triList[x].polyPlane.nx, triList[x].polyPlane.ny, triList[x].polyPlane.nz );
			glVertex3f( triList[x].polyVerts[y].x, triList[x].polyVerts[y].y, triList[x].polyVerts[y].z );
		}
		glEnd();
	}

	glPopMatrix();
}