#pragma once

#include "oglapp.h"

#include "gltga.h"
#include "vector.h"


class CMyOpenGLApp : public COGLApplication
{
private:

	GLuint m_vtxprog_gen;		// General transform
	GLuint rclist_passthrough;	// Pass the texture right on through into the output

	GLuint wall;

	glTga_t texture;

protected:
	
	OGL_ERRORS InitialSceneInit( void );                  
	OGL_ERRORS InitObjects( void );                        
	OGL_ERRORS RestoreObjects( void );                     
	OGL_ERRORS ResizeOpenGLScene( int width, int height ); 
	OGL_ERRORS NextFrame( void );		                 
	OGL_ERRORS Render( void );                             
	OGL_ERRORS CleanupObjects( void );

	// Returns the normalized cube map texture object ready to go
	GLuint SetupNormalCubeMap( const int size );
	
	void DrawRoom( void );
	void DrawBox( void );

public:

	CMyOpenGLApp();
};

// General transform program for a 2-TMU GPU
const char vtxprog_gen[] =
"!!VP1.0							\n"
"DP4 o[HPOS].x, c[0], v[OPOS];		\n"
"DP4 o[HPOS].y, c[1], v[OPOS];		\n"
"DP4 o[HPOS].z, c[2], v[OPOS];		\n"
"DP4 o[HPOS].w, c[3], v[OPOS];		\n"

"MOV o[COL0], v[COL0];				\n"
"MOV o[COL1], v[COL1];				\n"
"MOV o[TEX0], v[TEX0];				\n"
"MOV o[TEX1], v[TEX1];				\n"

"END								\n";


// Converts a normal [-1,1] to an rgb value of [0,255]
// -1 <= n.(x,y,z) <= 1  to:
//		-1*127+127 <= n.(x,y,z) <= 1*127+127
//				Evaluates to: (with rounding errors)
//		 0 <= n.(x,y,z) <= 255
void vec_to_rgb( VECTOR &v, unsigned char *ip );

extern const GLuint g_cubeMapDirections[6];
