#include "matrix.h"

/*
	    0   1   2    3      <- column
	0 | m1  m5  m9   m13 |
	1 | m2  m6  m10  m14 |
	2 | m3  m7  m11  m15 |
	3 | m4  m8  m12  m16 |

    ^- row
*/
// Remember: Specify matrix by Column x Row

void CMatrix::Multiply( const CMatrix &lhs, const CMatrix &rhs )
{
#define R rhs.m
#define L lhs.m

	m[0][0] = R[0][0] * L[0][0] + R[0][1] * L[1][0] + R[0][2] * L[2][0] + R[0][3] * L[3][0];
	m[1][0] = R[1][0] * L[0][0] + R[1][1] * L[1][0] + R[1][2] * L[2][0] + R[1][3] * L[3][0];
	m[2][0] = R[2][0] * L[0][0] + R[2][1] * L[1][0] + R[2][2] * L[2][0] + R[2][3] * L[3][0];
	m[3][0] = R[3][0] * L[0][0] + R[3][1] * L[1][0] + R[3][2] * L[2][0] + R[3][3] * L[3][0];

	m[0][1] = R[0][0] * L[0][1] + R[0][1] * L[1][1] + R[0][2] * L[2][1] + R[0][3] * L[3][1];
	m[1][1] = R[1][0] * L[0][1] + R[1][1] * L[1][1] + R[1][2] * L[2][1] + R[1][3] * L[3][1];
	m[2][1] = R[2][0] * L[0][1] + R[2][1] * L[1][1] + R[2][2] * L[2][1] + R[2][3] * L[3][1];
	m[3][1] = R[3][0] * L[0][1] + R[3][1] * L[1][1] + R[3][2] * L[2][1] + R[3][3] * L[3][1];

	m[0][2] = R[0][0] * L[0][2] + R[0][1] * L[1][2] + R[0][2] * L[2][2] + R[0][3] * L[3][2];
	m[1][2] = R[1][0] * L[0][2] + R[1][1] * L[1][2] + R[1][2] * L[2][2] + R[1][3] * L[3][2];
	m[2][2] = R[2][0] * L[0][2] + R[2][1] * L[1][2] + R[2][2] * L[2][2] + R[2][3] * L[3][2];
	m[3][2] = R[3][0] * L[0][2] + R[3][1] * L[1][2] + R[3][2] * L[2][2] + R[3][3] * L[3][2];

	m[0][3] = R[0][0] * L[0][3] + R[0][1] * L[1][3] + R[0][2] * L[2][3] + R[0][3] * L[3][3];
	m[1][3] = R[1][0] * L[0][3] + R[1][1] * L[1][3] + R[1][2] * L[2][3] + R[1][3] * L[3][3];
	m[2][3] = R[2][0] * L[0][3] + R[2][1] * L[1][3] + R[2][2] * L[2][3] + R[2][3] * L[3][3];
	m[3][3] = R[3][0] * L[0][3] + R[3][1] * L[1][3] + R[3][2] * L[2][3] + R[3][3] * L[3][3];
}

void CMatrix::Multiply( const vector_t &src, vector_t &dest )
{
	dest.x = m[0][0] * src.x + m[0][1] * src.y + m[0][2] * src.z + m[0][3] * src.w;
	dest.y = m[1][0] * src.x + m[1][1] * src.y + m[1][2] * src.z + m[1][3] * src.w;
	dest.z = m[2][0] * src.x + m[2][1] * src.y + m[2][2] * src.z + m[2][3] * src.w;
	dest.w = m[3][0] * src.x + m[3][1] * src.y + m[3][2] * src.z + m[3][3] * src.w;
}

void CMatrix::Rotate( const float rot, const float x, const float y, const float z )
{
	CMatrix mat;

	float rx, ry, rz;

	rx = x; ry = y; rz = z;

	if( rx >= 0.0f ) rx = 1.0f;
	if( ry >= 0.0f ) ry = 1.0f;
	if( rz >= 0.0f ) rz = 1.0f;
	if( rx < 0.0f ) rx = 0.0f;
	if( ry < 0.0f ) ry = 0.0f;
	if( rz < 0.0f ) rz = 0.0f;

	float c  = (float)cos( ( rot * 3.14159f ) / 180.0f );
	float s  = (float)sin( ( rot * 3.14159f ) / 180.0f );
	float mc = 1.0f - c;
	
	mat.m[0][0] = ( rx * rx * mc ) + c;
	mat.m[0][1] = ( ry * rx * mc ) + ( rz * s );
	mat.m[0][2] = ( rx * rz * mc ) - ( ry * s );

	mat.m[1][0] = ( rx * ry * mc ) - ( rz * s );
	mat.m[1][1] = ( ry * ry * mc ) + c;
	mat.m[1][2] = ( ry * rz * mc ) + ( rx * s );

	mat.m[2][0] = ( rx * rz * mc ) + ( ry * s );
	mat.m[2][1] = ( ry * rz * mc ) - ( rx * s );
	mat.m[2][2] = ( rz * rz * mc ) + c;

	Multiply( *this, mat );
}

void CMatrix::Rotate( const float rot, const vector_t &v )
{
	CMatrix mat;

	float rx, ry, rz;

	rx = v.x; ry = v.y; rz = v.z;

	if( rx >= 0.0f ) rx = 1.0f;
	if( ry >= 0.0f ) ry = 1.0f;
	if( rz >= 0.0f ) rz = 1.0f;
	if( rx < 0.0f ) rx = 0.0f;
	if( ry < 0.0f ) ry = 0.0f;
	if( rz < 0.0f ) rz = 0.0f;

	float c  = (float)cos( ( rot * 3.14159f ) / 180.0f );
	float s  = (float)sin( ( rot * 3.14159f ) / 180.0f );
	float mc = 1.0f - c;
	
	mat.m[0][0] = ( rx * rx * mc ) + c;
	mat.m[0][1] = ( ry * rx * mc ) + ( rz * s );
	mat.m[0][2] = ( rx * rz * mc ) - ( ry * s );

	mat.m[1][0] = ( rx * ry * mc ) - ( rz * s );
	mat.m[1][1] = ( ry * ry * mc ) + c;
	mat.m[1][2] = ( ry * rz * mc ) + ( rx * s );

	mat.m[2][0] = ( rx * rz * mc ) + ( ry * s );
	mat.m[2][1] = ( ry * rz * mc ) - ( rx * s );
	mat.m[2][2] = ( rz * rz * mc ) + c;

	Multiply( *this, mat );
}

void CMatrix::Translate( const float x, const float y, const float z )
{
	CMatrix mat;

	mat.m[3][0] = x;
	mat.m[3][1] = y;
	mat.m[3][2] = z;

	Multiply( *this, mat );
}

void CMatrix::Translate( const vector_t &v )
{
	CMatrix mat;

	mat.m[3][0] = v.x;
	mat.m[3][1] = v.y;
	mat.m[3][2] = v.z;

	Multiply( *this, mat );
}

bool CMatrix::Inverse( CMatrix &inverse )
{
	int i, j, k, swap;

	double t;
	double temp[4][4];

    for( i = 0; i < 4; ++i )
		for( j = 0; j < 4; ++j )
			temp[i][j] = ma[i * 4 + j]; 
      
   inverse.Identity();

   for( i = 0; i < 4; i++ )
   {
		// Look for largest element in column
		swap = i;
		for( j = i + 1; j < 4; ++j )
			if( fabs( temp[j][i] ) > fabs( temp[i][i] ) )
				swap = j;      

		if( swap != i )
		{
			// Swap rows.
			for( k = 0; k < 4; ++k ) 
			{
				t = temp[i][k];
				temp[i][k] = temp[swap][k];
				temp[swap][k] = t;

				t = inverse.ma[i * 4 + k];
				inverse.ma[i * 4 + k] = inverse.ma[swap * 4 + k];
				inverse.ma[swap * 4 + k] = t;
			}
		}

		// No non-zero pivot.  The matrix is singular, which shouldn't
		// happen.  This means the user gave us a bad matrix.
		if (temp[i][i] == 0)
			return false;
	   
		t = temp[i][i];

		for( k = 0; k < 4; ++k )
		{
			temp[i][k]            /= t;
			inverse.ma[i * 4 + k] /= t;
		}

		for( j = 0; j < 4; j++ )
			if( j != i )
			{
				t = temp[j][i];

				for( k = 0; k < 4; k++ )
				{
					temp[j][k]            -= temp[i][k] * t;
					inverse.ma[j * 4 + k] -= inverse.ma[i * 4 + k] * t;
				}
			}
		
   }
   
   return true;
}