#include <math.h>
#include <memory.h>

#include "vector.h"

#pragma once

class CMatrix
{
public:
	
	CMatrix() { Identity(); }


	double *GetMatrixSingleArray( void ) { return ma; }
	double  GetMatrixSingleArrayValue( const unsigned int ndx ) { return ma[ndx]; }

	void GetMatrix16( double _ma[] )    { memcpy( _ma, ma, sizeof( float[16] ) ); }
	void GetMatrix4x4( double _m[][4] ) { memcpy( _m, m, sizeof( float[4][4] ) ); }
	
	// OpenGL uses ColxRow b/c OpenGL matrices are stroed in Column Major
	inline double GetMatrixColRow( const unsigned int col, const unsigned int row ) const 
			{ return m[col][row]; }

	// D3D uses RowxCol which is the standard C/C++ way; stored in Row Major
	inline double GetMatrixRowCol( const unsigned int row, const unsigned int col ) const
			{ return m[row][col]; }

	inline void Store( const CMatrix &mat );
	inline void StoreDouble16( const double mat[] );
	inline void StoreDouble4x4( const double mat[][4] );
	inline void Identity( void );
	inline void Transpose( CMatrix &mat );
	inline void Add( const CMatrix &src );
	inline void Subtract( const CMatrix &src );
		
	void Multiply( const CMatrix &lhs, const CMatrix &rhs );
	//void Multiply( const CMatrix &lhs, const vector_t &rhsVec );
	//void Multiply( const vector_t &lhsVec, const CMatrix &rhs );
	void Multiply( const vector_t &src, vector_t &dest );

	void Rotate( const float rot, const float x, const float y, const float z );
	void Rotate( const float rot, const vector_t &v );

	void Translate( const float x, const float y, const float z );
	void Translate( const vector_t &v );

	bool Inverse( CMatrix &inverse );


	vector_t operator[]( int i ) 
		{ return vector_t( ma[i], ma[i + 4], ma[i + 8], ma[i + 12] ); }
	
private:

	union
	{
		double m[4][4];
		double ma[16];
	};	
};


void CMatrix::Store( const CMatrix &mat )
{
	for( int x = 0; x < 16; ++x )
		ma[x] = mat.ma[x];
}

void CMatrix::StoreDouble16( const double mat[] )
{
	for( int x = 0; x < 16; ++x )
		ma[x] = mat[x];
}

void CMatrix::StoreDouble4x4( const double mat[][4] )
{
	for( int x = 0; x < 4; ++x )
		for( int y =0 ; y < 4; ++y )
			m[x][y] = mat[x][y];
}

void CMatrix::Identity( void )
{
	for( int x = 0; x < 16; ++x )
	{
		if( x == 0 || ( x % 5 == 0 ) )
			ma[x] = 1.0;
		else
			ma[x] = 0.0;
	}
}

void CMatrix::Transpose( CMatrix &mat )
{
	mat.m[0][0] = m[0][0];
	mat.m[0][1] = m[1][0];
	mat.m[0][2] = m[2][0];
	mat.m[0][3] = m[3][0];

	mat.m[1][0] = m[0][1];
	mat.m[1][1] = m[1][1];
	mat.m[1][2] = m[2][1];
	mat.m[1][3] = m[3][1];

	mat.m[2][0] = m[0][2];
	mat.m[2][1] = m[1][2];
	mat.m[2][2] = m[2][2];
	mat.m[2][3] = m[3][2];

	mat.m[3][0] = m[0][3];
	mat.m[3][1] = m[1][3];
	mat.m[3][2] = m[2][3];
	mat.m[3][3] = m[3][3];
}

void CMatrix::Add( const CMatrix &src )
{
	for( int x = 0; x < 16; ++x )
		ma[x] += src.ma[x];
}

void CMatrix::Subtract( const CMatrix &src )
{
	for( int x = 0; x < 16; ++x )
		ma[x] -= src.ma[x];
}

