//////////////////////////////////////////////////////////////////////////////////////////////////
//		My 2d and 3d VECTOR class
//
//		Copyright by Thomas Grimes
//		6-23-00
//
//			You may use this code only with my permission and must not claim it as your own.
//			You may use this code to learn from and write your own but no copy and paste from mine
//			on to yours.
//
//		Filename: vector.cpp
//			Implements the 2d and 3d vector class
//////////////////////////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <windows.h>
#include <math.h>

#include "vector.h"

CVector::~CVector() {};

//
/// Addition
//
CVector CVector::operator + ( const CVector& vec )	
{
	return CVector( this->x + vec.x, this->y + vec.y, this->z + vec.z, this->w + vec.w );
}

//
/// Subtraction
//
CVector CVector::operator - ( const CVector& vec )		
{
	return CVector( this->x - vec.x, this->y - vec.y, this->z - vec.z, this->w - vec.w );
}

//
/// Multiplication
//
CVector CVector::operator * ( const CVector& vec )	
{
	return CVector( this->x * vec.x, this->y * vec.y, this->z * vec.z, this->w * vec.w );
}

//
/// Division
//
CVector CVector::operator / ( const CVector& vec )	
{
	return CVector( this->x / vec.x, this->y / vec.y, this->z / vec.z, this->w / vec.w );
}

//
/// Equal
//
CVector & CVector::operator = ( const CVector& vec )			
{
	if(this == &vec)		// Check addresses, if equal
		return *this;	// return 1st operand

	this->x = vec.x;
	this->y = vec.y;
	this->z = vec.z;
	this->w = vec.w;

	return *this;
}

//
/// Plus Equal
//
CVector & CVector::operator += ( const CVector& vec ) 			
{
	this->x = this->x + vec.x;
	this->y = this->y + vec.y;
	this->z = this->z + vec.z;
	this->w = this->w + vec.w;

	return *this;
}

//
/// Minus Equal
//
CVector & CVector::operator -= ( const CVector& vec ) 		
{
	this->x = this->x - vec.x;
	this->y = this->y - vec.y;
	this->z = this->z - vec.z;
	this->w = this->w - vec.w;

	return *this;
}

//
/// Divide Equal
//
CVector & CVector::operator /= ( const CVector& vec ) 		
{
	this->x = this->x / vec.x;
	this->y = this->y / vec.y;
	this->z = this->z / vec.z;
	this->w = this->w / vec.w;

	return *this;
}

//
/// Times Equal
//
CVector & CVector::operator *= ( const CVector& vec ) 	
{
	this->x = this->x * vec.x;
	this->y = this->y * vec.y;
	this->z = this->z * vec.z;
	this->w = this->w * vec.w;

	return *this;
}



//
/// VectorLength
//
inline float VectorLength( CVector& vec ) 
{
	float length = 0.0f;

	length = (float)sqrt( (pow( vec.x, 2 ) + pow( vec.y, 2 ) + pow( vec.z, 2 ) + pow( vec.w, 2)) );

	return length;
}

float VectorLength_2D( CVector& vec )  	
{
	float length = 0.0f;

	length = (float)sqrt( (pow( vec.x, 2 ) + pow( vec.y, 2 )) );

	return length;
}

//
/// DotProduct - Finds Dot product between vec1 and vec1, returns the cosine of the angle between
///				 vec1 and vec2.
//
inline float DotProduct( const CVector& vec1 , const CVector& vec2 ) 	
{
	//float dot = 0.0f;
	//dot = ( vec1.x * vec2.x ) + ( vec1.y * vec2.y ) + ( vec1.z * vec2.z ) + ( vec1.w * vec2.w );
	//return dot;

	return ( ( vec1.x * vec2.x ) + ( vec1.y * vec2.y ) + ( vec1.z * vec2.z ) + ( vec1.w * vec2.w ) );
}

float DotProduct_2D( CVector& vec1, CVector& vec2 )  		
{
	float dot = 0.0f;

	dot = ( vec1.x * vec2.x ) + ( vec1.y * vec2.y );

	return dot;
}

//
/// Normalize (direction vector which has a size of 1.0)
//
inline void Normalize( CVector& vec )  
{
	float length = 1.0f / VectorLength( vec );

	if( length == 0 )
		return;

	vec.x *= length;
	vec.y *= length;
	vec.z *= length;
	vec.w *= length;
}

void Normalize_2D( CVector& vec )  
{
	float length = 1.0f / VectorLength_2D( vec );

	if( length == 0 )
		return;

	vec.x *= length;
	vec.y *= length;
}

//
/// Clear Vector
//
inline void ClearVector( CVector& vec )  
{
	vec.x = vec.y = vec.z = vec.w = 0;
}

void ClearVector_2D( CVector& vec )  
{
	vec.x = vec.y = 0;
}

//
/// CrossProduct 
//
inline CVector CrossProduct( const CVector& vec1, const CVector& vec2 )  
{
	CVector cross;
	cross.x = (vec1.y * vec2.z) - (vec1.z * vec2.y);
	cross.y = (vec1.z * vec2.x) - (vec1.x * vec2.z);
	cross.z = (vec1.x * vec2.y) - (vec1.y * vec2.x);

	return cross;
}

// 
/// Scale Vector by another vector
//
inline void ScaleVector( CVector& vec1, CVector& vec2 )  
{
	vec1.x += vec2.x;
	vec1.y += vec2.y;
	vec1.z += vec2.z;
	vec1.w += vec2.w;
}

void ScaleVector_2D( CVector& vec1, CVector& vec )  
{
	vec1.x += vec.x;
	vec1.y += vec.y;
}

//
/// Scale Vector by float value
//
inline void ScaleVectorByValue( CVector& vec, float scaleFactor )  
{
	vec.x += scaleFactor;
	vec.y += scaleFactor;
	vec.z += scaleFactor;
	vec.w += scaleFactor;
}

void ScaleVectorByValue_2D( CVector& vec, float scaleFactor )  
{
	vec.x += scaleFactor;
	vec.y += scaleFactor;
}

//
/// Clamp Vector
//
inline void ClampVector( CVector& vec, int min, int max )  
{
	if( vec.x > max || vec.x < min )
		vec.x = 0;
	if( vec.y > max || vec.y < min )
		vec.y = 0;
	if( vec.z > max || vec.z < min )
		vec.z = 0;
	if( vec.w > max || vec.w < min )
		vec.w = 0;
}

void ClampVector_2D( CVector& vec, int min, int max ) 
{
	if( vec.x > max || vec.x < min )
		vec.x = 0;
	if( vec.y > max || vec.y < min )
		vec.y = 0;
}

//
/// Copy to Array
//
// Becarefull with these two functions
inline void CopyToArray( CVector& vec, float* array ) 
{
	array[0] = vec.x;
	array[1] = vec.y;
	array[2] = vec.z;
	array[3] = vec.w;
}

void CopyToArray_2D( CVector& vec, float* array )
{
	array[0] = vec.x;
	array[1] = vec.y;
}

inline void CopyPointerToValue( CVector &vec, const CVector *vec2 )
{
	vec.x = vec2->x;
	vec.y = vec2->y;
	vec.z = vec2->z;
}

//
/// TangentBasis
//
void TangentBasis( const VECTOR &vertex, const VECTOR &vertex2, const VECTOR &vertex3, 
				   const float texcoords[], const float texcoords2[], const float texcoords3[], 
				   const VECTOR &polynormal, VECTOR &tangent, VECTOR &binormal, VECTOR &normal )
{
	VECTOR txb;
	VECTOR v1( vertex2.x - vertex.x, texcoords2[0] - texcoords[0], texcoords2[1] - texcoords[1] );
	VECTOR v2( vertex3.x - vertex.x, texcoords3[0] - texcoords[0], texcoords3[1] - texcoords[1] );

	
	txb = CrossProduct( v1, v2 );
	
	if( fabs( txb.x ) > EPSILON )
	{
		tangent.x  = -txb.y / txb.x;
		binormal.x = -txb.z / txb.x;
	}

	v1.x = vertex2.y - vertex.y;
	v2.x = vertex3.y - vertex.y;

	txb = CrossProduct( v1, v2 );
	
	if( fabs( txb.x ) > EPSILON )
	{
		tangent.y  = -txb.y / txb.x;
		binormal.y = -txb.z / txb.x;
	}

	v1.x = vertex2.z - vertex.z;
	v2.x = vertex3.z - vertex.z;

	txb = CrossProduct( v1, v2 );
	
	if( fabs( txb.x ) > EPSILON )
	{
		tangent.z  = -txb.y / txb.x;
		binormal.z = -txb.z / txb.x;
	}

	Normalize( tangent );
	Normalize( binormal );

	// Make a normal based on the tangent and binormal b/c it may be different than the poly's
	// normal, this normal being computed here is better
	normal = CrossProduct( tangent, binormal );
	Normalize( normal );

	// Make tangent space vectors orthogonal by recomputing the binormal with the corrected 
	// tangent space normal.
	binormal = CrossProduct( tangent, normal );
	Normalize( binormal );

	// Check the generated normal from the tangent and binormal vectors against the polygon normal
	if( DotProduct( normal, polynormal ) < 0.0f )
	{
		normal.x = -normal.x;
		normal.y = -normal.y;
		normal.z = -normal.z; 
	}
		
}


