//////////////////////////////////////////////////////////////////////////////////////////////////
//		My 2d and 3d VECTOR class
//
//		Copyright by Thomas Grimes
//		6-23-00
//
//			You may use this code only with my permission and must not claim it as your own.
//			You may use this code to learn from and write your own but no copy and paste from mine
//			on to yours.
//
//		Filename: vector.h
//			Defines the 2d and 3d vector class
//////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef VECTOR_H
#define VECTOR_H

// Single dim vector
typedef float vec_t;

#define EPSILON 0.00001f	// 1.0 x 10^(-5)

class CVector
{
	public:
		// Default constructor
		CVector() { x = y = z = w = normal[0] = normal[1] = normal[2] = 0; };
		// 2D Vector Constructor
		CVector( float vectorX, float vectorY ) { x = vectorX; y = vectorY; z = w = 0; };
		// 3D Vector Constructor
		CVector( float vectorX, float vectorY, float vectorZ ) { x = vectorX; y = vectorY; z = vectorZ; w = 0; };
		// 4D Vector
		CVector( float vectorX, float vectorY, float vectorZ, float vectorW ) {  x = vectorX; y = vectorY; z = vectorZ; w = vectorW; } 

		// Destructor
		~CVector();

		// Overloading
		// NOTE: OVERLOADED OPERATORS ONLY WORK WITH A 3D VECTOR !!
		CVector operator + ( const CVector& vec );	// Addition       
		CVector operator - ( const CVector& vec );	// Subtraction    
		CVector operator * ( const CVector& vec );	// Multiplication 
		CVector operator / ( const CVector& vec );	// Division       
				
		CVector& operator =  ( const CVector& vec );	// Equal		  
		CVector& operator += ( const CVector& vec );	// Plus & Equal	  
		CVector& operator -= ( const CVector& vec );	// Minus & Equal  
		CVector& operator /= ( const CVector& vec );	// Divide & Equal 
		CVector& operator *= ( const CVector& vec );	// Times & Equal  
				
		// I may use these two
		//bool operator==(const CVector& vec) const;		// Equal Equal
		//bool operator!=(const CVector& vec) const;		// Not Equal

			
		// Vector Coords
		vec_t x; 
		vec_t y;
		vec_t z;
		vec_t w;

		// Normal
		float normal[3];

		// Linked List of vectors
		//CVector* next;
};

typedef CVector VECTOR;
typedef CVector vector_t;



inline float VectorLength( CVector& vec );
float VectorLength_2D( CVector& vec );
		
inline float DotProduct( const CVector& vec1 , const CVector& vec2 );
float DotProduct_2D( CVector& vec1 , CVector& vec2 );
		
inline void Normalize( CVector& vec );					
void Normalize_2D( CVector& vec );
		// Set to 0
inline void ClearVector( CVector& vec );					
void ClearVector_2D( CVector& vec );
		// Find CP of vec1 and vec2 and return the result
inline CVector CrossProduct( const CVector& vec1, const CVector& vec2 );	
		
inline void ScaleVector( CVector& vec1 , CVector& vec2 );		
void ScaleVector_2D( CVector& vec1, CVector& vec2 );
		
inline void ScaleVectorByValue( CVector& vec, float scaleFactor );
void ScaleVectorByValue_2D( CVector& vec, float scaleFactor );
	
inline void ClampVector( CVector& vec, int min, int max );		
void ClampVector_2D( CVector& vec, int min, int max );
		
inline void CopyToArray( CVector& vec, float* array );				
void CopyToArray_2D( CVector& vec, float* array );

inline void CopyPointerToValue( CVector &vec, const CVector *vec2 );

// Fast vector funcs
// Will using 'register' make any difference?
inline void VectorAdd( vector_t *result, const vector_t &vec1, const vector_t &vec2 )
{
	result->x = vec1.x + vec2.x;
	result->y = vec1.y + vec2.y;
	result->z = vec1.z + vec2.z;
}

inline void VectorAddFloat( vector_t *result, const vector_t &vec, const float &val )
{
	result->x = vec.x + val;
	result->y = vec.y + val;
	result->z = vec.z + val;
}

inline void VectorAddFloatConst( vector_t *result, const vector_t &vec, const float val )
{
	result->x = vec.x + val;
	result->y = vec.y + val;
	result->z = vec.z + val;
}

inline void VectorSubtract( vector_t *result, const vector_t &vec1, const vector_t &vec2 )
{
	result->x = vec1.x - vec2.x;
	result->y = vec1.y - vec2.y;
	result->z = vec1.z - vec2.z;
}

inline void VectorSubFloat( vector_t *result, const vector_t &vec, const float &val )
{
	result->x = vec.x - val;
	result->y = vec.y - val;
	result->z = vec.z - val;
}

inline void VectorSubFloatConst( vector_t *result, const vector_t &vec, const float val )
{
	result->x = vec.x - val;
	result->y = vec.y - val;
	result->z = vec.z - val;
}

inline void VectorMultiply( vector_t *result, const vector_t &vec1, const vector_t &vec2 )
{
	result->x = vec1.x * vec2.x;
	result->y = vec1.y * vec2.y;
	result->z = vec1.z * vec2.z;
}

inline void VectorMulFloat( vector_t *result, const vector_t &vec, const float &val )
{
	result->x = vec.x * val;
	result->y = vec.y * val;
	result->z = vec.z * val;
}

inline void VectorMulFloatConst( vector_t *result, const vector_t &vec, const float val )
{
	result->x = vec.x * val;
	result->y = vec.y * val;
	result->z = vec.z * val;
}

// MAD will use function overloading
inline void VectorMAD( vector_t *result, const vector_t &vec1, const vector_t &vec2, const vector_t &vec3 )
{
	result->x = vec1.x * vec2.x + vec3.x;
	result->y = vec1.y * vec2.y + vec3.y;
	result->z = vec1.z * vec2.z + vec3.z;
}

inline void VectorMAD( vector_t *result, const vector_t &vec1, const vector_t &vec2, const float &val )
{
	result->x = vec1.x * vec2.x + val;
	result->y = vec1.y * vec2.y + val;
	result->z = vec1.z * vec2.z + val;
}

inline void VectorMAD( vector_t *result, const vector_t &vec1, const vector_t &vec2, const float val )
{
	result->x = vec1.x * vec2.x + val;
	result->y = vec1.y * vec2.y + val;
	result->z = vec1.z * vec2.z + val;
}

inline void VectorMAD( vector_t *result, const vector_t &vec1, const float val1, const vector_t &vec2 )
{
	result->x = vec1.x * val1 + vec2.x;
	result->y = vec1.y * val1 + vec2.y;
	result->z = vec1.z * val1 + vec2.z;
}

inline void VectorMAD( vector_t *result, const vector_t &vec1, const float &val1, const float &val2 )
{
	result->x = vec1.x * val1 + val2;
	result->y = vec1.y * val1 + val2;
	result->z = vec1.z * val1 + val2;
}

inline void VectorMAD( vector_t *result, const vector_t &vec1, const float &val1, const float val2 )
{
	result->x = vec1.x * val1 + val2;
	result->y = vec1.y * val1 + val2;
	result->z = vec1.z * val1 + val2;
}

inline void VectorMAD( vector_t *result, const vector_t &vec1, const float val1, const float &val2 )
{
	result->x = vec1.x * val1 + val2;
	result->y = vec1.y * val1 + val2;
	result->z = vec1.z * val1 + val2;
}

inline void VectorMAD( vector_t *result, const vector_t &vec1, const float val1, const float val2 )
{
	result->x = vec1.x * val1 + val2;
	result->y = vec1.y * val1 + val2;
	result->z = vec1.z * val1 + val2;
}

inline void VectorDivide( vector_t *result, const vector_t &vec1, const vector_t &vec2 )
{
	result->x = vec1.x / vec2.x;
	result->y = vec1.y / vec2.y;
	result->z = vec1.z / vec2.z;
}

inline void VectorDivFloat( vector_t *result, const vector_t &vec, const float &val )
{
	result->x = vec.x / val;
	result->y = vec.y / val;
	result->z = vec.z / val;
}

inline void VectorDivFloatConst( vector_t *result, const vector_t &vec, const float val )
{
	result->x = vec.x / val;
	result->y = vec.y / val;
	result->z = vec.z / val;
}

inline void VectorInvert( vector_t *result, const vector_t &vec )
{
	result->x = -vec.x;
	result->y = -vec.y;
	result->z = -vec.z;
}

inline void VectorMagnitude( vector_t *result, const vector_t &vec )
{
	( ( vec.x < 0 ) ? ( result->x = -vec.x ) : ( result->x = vec.x ) );
	( ( vec.y < 0 ) ? ( result->y = -vec.x ) : ( result->y = vec.x ) );
	( ( vec.z < 0 ) ? ( result->z = -vec.x ) : ( result->z = vec.x ) );	
}

////////////

void TangentBasis( const VECTOR &vertex, const VECTOR &vertex2, const VECTOR &vertex3, 
				   const float texcoords[], const float texcoords2[], const float texcoords3[], 
				   const VECTOR &polynormal, VECTOR &tangent, VECTOR &binormal, VECTOR &normal );

#endif